/*
 *	pid.c
 *
 *	process ID management
 */

#include "mhpopd.h"
#include "fileio.h"

static int NEAR pid_filename __P_((char *, ALLOC_T));


static int NEAR pid_filename(path, size)
char *path;
ALLOC_T size;
{
	uid_t uid;
	CONST char *subdir;
	char *cp;
	int n;

	cp = (piddir && *piddir == '/') ? piddir : DEFPIDDIR;
	subdir = nullstr;
	if ((uid = geteuid())) {
		cp = homepiddir;
		if (!cp || *cp != '/') {
			if (cp) subdir = cp;
			if (!(cp = pwd_gethome(uid))) return(-1);
		}
	}
	n = Xsnprintf(path, size, "%s", cp);
	if (Xmkdir(path, 0755) < 0) return(-1);
	if (subdir && *subdir) {
		n += Xsnprintf(&(path[n]), size - n, "/%s", subdir);
		if (Xmkdir(path, 0755) < 0) return(-1);
	}
	Xsnprintf(&(path[n]), size - n, "/%s.pid", myname);

	return(0);
}

int pid_check(VOID_A)
{
	XFILE *fp;
	p_id_t pid;
	char path[MAXPATHLEN], buf[MAXLONGWIDTH + 1 + 1];
	int n;

	if (pid_filename(path, sizeof(path)) < 0) return(-1);
	if (!(fp = Xfopen(path, "r", 0, XF_IGNORENOENT))) {
		if (errno == ENOENT) return(0);
		n = -1;
	}
	else {
		n = Xfread(buf, strsize(buf), fp);
		VOID_C Xfclose(fp);
	}

	if (n >= 0) {
		buf[n] = '\0';
		pid = atoi(buf);
		if (kill(pid, 0) >= 0) {
			ERROR0(("Daemon running already (pid=%d)", pid));
			return(-1);
		}
	}

	VOID_C Xunlink(path);

	return(0);
}

int pid_create(VOID_A)
{
	XFILE *fp;
	char path[MAXPATHLEN], buf[MAXLONGWIDTH + 1 + 1];
	int n;

	if (ischild || isinetd) return(0);
	if (pid_filename(path, sizeof(path)) < 0) return(-1);
	n = umask(022);
	fp = Xfopen(path, "w", O_EXCL, 0);
	VOID_C umask(n);
	if (!fp) {
		VOID_C Xunlink(path);
		return(-1);
	}

	n = Xsnprintf(buf, sizeof(buf), "%d\n", getpid());
	n = (Xfwrite(buf, n, fp) < n) ? -1 : 0;
	if (n >= 0) n = (Xfflush(fp) == EOF) ? -1 : 0;
	VOID_C Xfclose(fp);

	if (n < 0) VOID_C Xunlink(path);

	return(n);
}

VOID pid_unlink(VOID_A)
{
	char path[MAXPATHLEN];

	if (ischild || isinetd) return;
	if (pid_filename(path, sizeof(path)) < 0) return;

	VOID_C Xunlink(path);
}
