/*
 *	output.c
 *
 *	Output LIPS Command Sequence
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "lipsf.h"
#include "var.h"
#if 	MSDOS
#include <dos.h>
#include <direct.h>
#else
#include <sys/param.h>
#endif

#define	FF	0x0c	/* \014	*/
#define	SO	0x0e	/* \016	*/
#define	SI	0x0f	/* \017	*/
#define	ESC	0x1b	/* \033	*/

#if	MSDOS
#define	MAXPATHLEN	1024
#define	getwd(path)	getcwd(path, MAXPATHLEN);
#else
#define	lprintf	fprintf
#define	lputc	fputc
#endif

#define kanjiin(stream)		bufprintf(stream, "\033n")
#define kanjiout(stream)	bufputc(SI, stream)

int kanjiflag;
int kanjimode;
int kanjicode = DEF_KCODE;
int buffer;
struct stat filebuf;
char fullpathbody[MAXPATHLEN];
char filenamebody[MAXPATHLEN];

extern FILE *infile;
extern int column;
extern int line;
extern int logicalline;
extern long totalline;
extern int lastch;

char *strdup2();
VOID getfileinfo();
static VOID ungetchar();
int iskanjiin();
int iskanjiout();
VOID output();
VOID resetcursor();
VOID frontprocess();
VOID backprocess();

extern VOID_P malloc();
extern VOID free();

extern VOID error();
extern int getarg();
extern int bufputc();
extern VOID printout();
extern int filterputc();
#if	MSDOS
extern int lputc(int, FILE *);
extern int lprintf(FILE *, u_char *, ...);
extern int bufprintf(FILE *, u_char *, ...);
#else
extern int bufprintf();
#endif


char *strdup2(str)
char *str;
{
	char *tmp;

	if (!str) return(NULL);
	if (!(tmp = (char *)malloc(strlen(str) + 1))) error(NULL);
	strcpy(tmp, str);
	return(tmp);
}

VOID getfileinfo(file, fp)
char *file;
FILE *fp;
{
	char cwd[MAXPATHLEN];
	char *tmp, *tmp2, *cp;
#if	MSDOS
	int drive, curdrive, dammy;
#endif

	if (file) {
		if (!(tmp = tmp2 = strdup2(file))) error(-1);
#if	MSDOS
		_dos_getdrive(&curdrive);
		if (cp = strchr(tmp, ':')) {
			tmp2 = cp + 1;
			drive = *tmp;
			if (drive >= 'a' && drive <= 'z') drive += 'A' - 'a';
			drive -= 'A' - 1;
			_dos_setdrive(drive, &dammy);
		}
#endif
		getwd(cwd);
		cp = strrchr(tmp2, PATHDELIM);
		if (!cp) strcpy(filenamebody, tmp2);
		else {
			strcpy(filenamebody, cp + 1);
			if (cp == tmp2) cp++;
			*cp = '\0';
			if (chdir(tmp2)) error(-1);
		}
		getwd(fullpathbody);
		chdir(cwd);
#if	MSDOS
		_dos_setdrive(curdrive, &dammy);
#endif
		free(tmp);
	}
	else {
		strcpy(filenamebody, "( STDIN )");
		*fullpathbody = '\0';
	}
	if (istoupper) {
		for (cp = filenamebody; *cp; cp++)
			if (*cp >= 'a' && *cp <= 'z') *cp += 'A' - 'a';
		for (cp = fullpathbody; *cp; cp++)
			if (*cp >= 'a' && *cp <= 'z') *cp += 'A' - 'a';
	}
	if (istolower) {
		for (cp = filenamebody; *cp; cp++)
			if (*cp >= 'A' && *cp <= 'Z') *cp += 'a' - 'A';
		for (cp = fullpathbody; *cp; cp++)
			if (*cp >= 'A' && *cp <= 'Z') *cp += 'a' - 'A';
	}

	if (fstat(fileno(fp), &filebuf)) error(-1);
}

static VOID ungetchar(c)
int c;
{
	if (c != EOF) ungetc(c, infile);
}

int iskanjiin(c)
int c;
{
	int c1, c2;

	if (c == ESC) {
		if ((c1 = fgetc(infile)) != '$') ungetchar(c1);
		else {
			if (strchr("@B", (c2 = fgetc(infile)))) return(1);
			else {
				ungetchar(c2);
				ungetchar(c1);
			}
		}
	}
	return(0);
}

int iskanjiout(c)
int c;
{
	int c1, c2;

	if (c == ESC) {
		if ((c1 = fgetc(infile)) != '(') ungetchar(c1);
		else {
			if (strchr("JHB", (c2 = fgetc(infile)))) return(1);
			else {
				ungetchar(c2);
				ungetchar(c1);
			}
		}
	}
	return(0);
}

VOID output(c, stream)
int c;
FILE *stream;
{
	int kcode;

	kcode = kanjicode;
	if (kcode != SJIS && kcode != EUC) kcode = DEF_KCODE;

	if (kanjimode == ASCII) {
		if (iskanjiin(c)) {
			kanjimode = JIS;
			kanjiflag = 2;
			kanjiin(stream);
			return;
		}
		else if (kcode == SJIS) {
			if (issjis1(c)) {
				kanjimode = SJIS;
				bufprintf(stream, "\033%%0");
				kanjiflag = 1;
				kanjiin(stream);
			}
			else if (iseuc(c)) {
				kanjimode = EUC;
				bufprintf(stream, "\033%%0");
				kanjiflag = 1;
				if (c != 0x8e) kanjiin(stream);
			}
		}
		else {
			if (iseuc(c)) {
				kanjimode = EUC;
				bufprintf(stream, "\033%%0");
				kanjiflag = 1;
				if (c != 0x8e) kanjiin(stream);
			}
			else if (issjis1(c)) {
				kanjimode = SJIS;
				bufprintf(stream, "\033%%0");
				kanjiflag = 1;
				kanjiin(stream);
			}
		}
	}
	else if (kanjimode == SJIS) {
		switch (kanjiflag) {
			case 0: if (issjis1(c)) {
					kanjiflag = 1;
					kanjiin(stream);
				}
				break;
			case 1: kanjiflag++;
				break;
			case 2: if (issjis1(c)) kanjiflag = 1;
				else {
					kanjiflag = 0;
					kanjiout(stream);
				}
				break;
			default:
				break;
		}
	}
	else if (kanjimode == EUC) {
		switch (kanjiflag) {
			case 0: if (iseuc(c)) {
					kanjiflag = 1;
					if (c != 0x8e) kanjiin(stream);
				}
				break;
			case 1: kanjiflag++;
				break;
			case 2: if (iseuc(c)) kanjiflag = 1;
				else {
					kanjiflag = 0;
					kanjiout(stream);
				}
				break;
			default:
				break;
		}
	}
	else {
		switch (kanjiflag) {
			case 0: if (iskanjiin(c)) {
					kanjiflag = 2;
					kanjiin(stream);
					return;
				}
				break;
			case 1: kanjiflag++;
				break;
			case 2: if (!iskanjiout(c)) kanjiflag = 1;
				else {
					kanjiflag = 0;
					kanjiout(stream);
					return;
				}
				break;
			default:
				break;
		}
	}

	if (c == '\t') {
		do filterputc(' ', stream);
		while (column % tabposition);
		return;
	}
	else if (c == '\n') {
		if (lastch != '\r') filterputc('\r', stream);
		line++;
		totalline++;
		logicalline++;
	}
	else if (c == '\f') {
		line++;
		totalline++;
	}
	filterputc(c, stream);
}

VOID resetcursor(stream, x, y)
FILE *stream;
int x, y;
{
	lprintf(stream, "\033[?6 I");
	lprintf(stream, "\033[11h");
	lprintf(stream, "\033[%ld;%ldf",
		((long)topmargin + (long)y) * 100 - 500,
		(long)leftmargin * 100 - 500);

	lprintf(stream, "\033[7 I\033[11h");
	lprintf(stream, "\033[%da", x);
	lprintf(stream, "\033[7 I\033[11h");
	if (lettersize < 63) lprintf(stream, "\033[%dk", 63 - lettersize);
	else lprintf(stream, "\033[%de", lettersize - 63);
	lprintf(stream, "\033[0;2t");
}

VOID frontprocess(stream)
FILE *stream;
{
	long tmp = (lipsmode == 3) ? 300 : 240;

	lprintf(stream, "\033%%@");
	lprintf(stream, "\033P%1d1;%d;1J\033\\", lipsmode, tmp);
	lprintf(stream, "\033<");
	lprintf(stream, "\033[%1dq", cassette);

	lprintf(stream, "\033[?6 I");
	lprintf(stream, "\033[%d", papersize + islandscape);
	if (papersize == 80) lprintf(stream, ";%ld;%ld",
		(long)paperhight * 100, (long)paperwidth * 100);
	lputc('p', stream);

	lprintf(stream, "\033(J");
	lprintf(stream, "\033)I");
	lprintf(stream, "\033$*,0");

	lprintf(stream, "\033[7 I");
	lprintf(stream, "\033[?%d K", (tmp * 100 * 2) / lettersize);
	lprintf(stream, "\033[%d C", lettersize);
	lprintf(stream, "\033[%dy", fontpattern);
	lprintf(stream, "\033n");
	lprintf(stream, "\033[?%d K", (tmp * 100) / lettersize);
	lprintf(stream, "\033[%d C", lettersize);
	lprintf(stream, "\033[%dy", fontpattern);
	lputc(SI, stream);

	lprintf(stream, "\033[%d S", isvirtical);
	lprintf(stream, "\033[?29%c", (isvirtical) ? 'h' : 'l');

	lprintf(stream, "\033[7 I");
	lprintf(stream, "\033[%d;%d G",
		linehight + lettersize, columnwidth + lettersize / 2);
	lprintf(stream, "\033n");
	lprintf(stream, "\033[%d;%d G",
		linehight + lettersize, (columnwidth + lettersize / 2) * 2);
	lputc(SI, stream);

	lprintf(stream, "\033[?23h");
	lprintf(stream, "\033[?25h");
	lprintf(stream, "\033[0%%u");

	lprintf(stream, "\033[%dv", printnum);

	switch (overlay) {
		case 1:
		case 2:	lprintf(stream, "\033[1;%1d;1&s", overlay - 1);
			break;
		case 3:
		case 4:	lprintf(stream, "\033[1;%1d&w", overlay - 3);
			break;
		case 5:	lprintf(stream, "\033[1;0;1&w");
			break;
		default:
			break;
	}

	resetcursor(stream, 0, 0);
}

VOID backprocess(stream)
FILE *stream;
{
	lputc(FF, stream);
	lprintf(stream, "\033P0J\033\\");
	if (overlay != 1 && overlay != 2) lprintf(stream, "\033c");
	lprintf(stream, "\033H");
}
