/*
 *	print.c
 *
 *	Print Out with Text Restructure
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include "lipsf.h"
#include "var.h"
#if 	MSDOS
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#define	FF	0x0c	/* \014	*/
#define	SO	0x0e	/* \016	*/
#define	SI	0x0f	/* \017	*/
#define	ESC	0x1b	/* \033	*/

#define MAXCOLS	512

#if	MSDOS
extern int lputc(int, FILE *);
extern int lprintf(FILE *, u_char *, ...);
#else
#define	lprintf	fprintf
#define	lputc	fputc
#endif

#define kanjiin(stream)		bufprintf(stream, "\033n")
#define kanjiout(stream)	bufputc(SI, stream)

static int halfline;
static u_char linebuf[MAXCOLS];
static int inrange;

int column;
int line;
int logicalline;
long totalline;
int step;
int page;
int lastch;
int nextch;
int linep;
int inheader;

extern FILE *infile;
extern int kanjiflag;
extern int kanjimode;
extern int buffer;
extern struct tm *nowtime;

static u_char *str2chr();
static VOID oneline();
static VOID header();
static VOID footer();
static VOID columnstep();
VOID bufflush();
int bufputc();
static int isoverline();
VOID linefeed();
VOID forcedlinefeed();
int foldputc();
VOID printout();

#if	MSDOS
int bufprintf(FILE *, u_char *, ...);
int foldprintf(FILE *, u_char *, ...);
#else
int bufprintf();
int foldprintf();
#endif

extern VOID error();
extern VOID output();
extern VOID resetcursor();
extern VOID stackvar();


static u_char *str2chr(s, c1, c2)
u_char *s;
int c1, c2;
{
	for (; *s; s += 2) if (*s == c1 && *(s + 1) == c2) return(s);
	return(NULL);
}

static VOID oneline(stream, str)
FILE *stream;
u_char *str;
{
	int i;
	int c, duplastch;
	int dupcolumn, duplinep;

	duplastch = lastch;
	dupcolumn = column;
	duplinep = linep;
	lastch = '\0';
	column = linep = 0;
	inheader = 1;
	for (i = 0;  c = str[i]; i++) {
		output(c, stream);
		lastch = c;
	}
	if (kanjiflag) kanjiout(stream);
	if (linep) bufputc('\r', stream);
	inheader = 0;
	lastch = duplastch;
	column = dupcolumn;
	linep = duplinep;
}

static VOID header(stream)
FILE *stream;
{
	if (headerstr) {
		oneline(stream, headerstr);
		lprintf(stream, "\033[?6 I");
		lprintf(stream, "\033[11h");
		lputc('\r', stream);
		lprintf(stream, "\033[%lde", (long)headermargin * 100);
		lprintf(stream, "\033[7 I");
		lprintf(stream, "\033[11h");
		lprintf(stream, "\033[%de", lettersize);
	}
	lprintf(stream, "\033[1;0;0x");
}

static VOID footer(stream)
FILE *stream;
{
	if (footerstr) {
		lprintf(stream, "\033[0;0;0x");
		lprintf(stream, "\033[7 I");
		lprintf(stream, "\033[11h");
		lprintf(stream, "\033[%de",
			(lettersize + linehight) * maxlines - linehight);
		lprintf(stream, "\033[?6 I");
		lprintf(stream, "\033[11h");
		lputc('\r', stream);
		lprintf(stream, "\033[%lde", (long)footermargin * 100);
		oneline(stream, footerstr);
	}
}

static VOID columnstep(stream)
FILE *stream;
{
	long tmp;

	if (maxcolumns) {
		tmp = maxcolumns;
		tmp *= columnwidth + lettersize / 2;
		tmp -= columnwidth;
		tmp += (dotprintwidth - tmp * maxsteps) / (maxsteps - 1);
	}
	else tmp = dotprintwidth / maxsteps;
	tmp = tmp * step;
	resetcursor(stream, (int)tmp, (headerstr && step) ? headermargin : 0);
}

VOID bufflush(stream)
FILE *stream;
{
	int i, j, indent;
	int flag;
	int col1, len1, col2, len2;
	u_char *tmp;

	flag = 0;

	if (inheader) {
		if (linebuf[linep - 1] == '\r') linep--;
		col1 = len1 = col2 = len2 = 0;

		if (tmp = (u_char *)memchr(linebuf, '\0', linep)) {
			col1 = tmp[1] + tmp[2] * 0x100;
			if (tmp = (u_char *)memchr(&tmp[3], '\0',
			linep - (int)(&tmp[3] - linebuf))) {
				col2 = tmp[1] + tmp[2] * 0x100;
				len1 = col2 - col1;
				len2 = column - col2;
			}
			else len1 = column - col1;
			col1 = (realmaxcolumns - len1) / 2 - col1;
			col2 = realmaxcolumns - len2 - col2 - col1;
			if (col2 < 0) {
				col1 += col2;
				if (col1 > 0) {
					col2 = col1 / 2;
					col1 -= col2;
				}
			}
		}
		for (i = 0; i < linep; i++) {
			if (linebuf[i]) lputc(linebuf[i], stream);
			else {
				switch (flag++) {
					case 0:	for (j = 0; j < col1; j++)
							lputc(' ', stream);
						break;
					case 1:	for (j = 0; j < col2; j++)
							lputc(' ', stream);
						break;
					default:
						break;
				}
				i += 2;
			}
		}
	}
	else {

		if (centering) indent = (maxcolumns - column) / 2;
		else if (flushright) indent = maxcolumns - column;
		else if (leftindent) indent = leftindent;
		else indent = 0;

		if (linep && linebuf[linep - 1] == '\f') {
			line = halfline = 0;
			linep--;
			flag = 1;
		}
		else if (maxlines && line >= maxlines - 1
		&& *linebuf != '\n') {
			if (linebuf[linep - 1] == '\r') linep--;
		}
		else if (maxlines && line >= maxlines) {
			line -= maxlines;
			halfline = 0;
			linep = 0;
			if (!step && inrange) footer(stream);
			if (maxsteps > 1) {
				if (++step >= maxsteps) step = 0;
				if (inrange) columnstep(stream);
			}
			if (maxsteps < 2 || !step ) {
				if (inrange) {
					lputc('\r', stream);
					lputc(FF, stream);
				}
				page++;
				inrange = (startpage <= 0 || page >= startpage)
					&& (endpage <= 0 || page <= endpage);
				if (inrange) header(stream);
			}
		}

		if (inrange) {
			if (*linebuf != '\n' && *linebuf != '\r')
				for (i = 0; i < indent; i++) lputc(' ', stream);
			for (i = 0; i < linep; i++) lputc(linebuf[i], stream);
		}

		if (flag) {
			if (!step && inrange) footer(stream);
			if (maxsteps > 1) {
				if (++step >= maxsteps) step = 0;
				if (inrange) columnstep(stream);
			}
			if (maxsteps < 2 || !step) {
				if (inrange) {
					lputc('\r', stream);
					lputc(FF, stream);
				}
				page++;
				inrange = (startpage <= 0 || page >= startpage)
					&& (endpage <= 0 || page <= endpage);
				if (inrange) header(stream);
			}
		}
		centering = flushright = 0;
	}
	linep = 0;
}

int bufputc(c, stream)
int c;
FILE *stream;
{
	linebuf[linep++] = c;
	if (linep >= MAXCOLS || c == '\r' || c == '\n' || c == '\f')
		bufflush(stream);
	else if (linep > 1 && linebuf[linep - 2] == '\033'
		&& strchr("KML", c)) bufflush(stream);
}

#ifndef	NOVSPRINTF
/*VARARGS1*/
#if	MSDOS
int bufprintf(FILE *stream, u_char *fmt, ...)
#else
int bufprintf(stream, fmt, va_alist)
FILE *stream;
u_char *fmt;
va_dcl
#endif
{
	va_list args;
	int i, len;
	u_char buf[MAXCOLS];

#if	MSDOS
	va_start(args, fmt);
#else
	va_start(args);
#endif

	len = vsprintf(buf, fmt, args);
#else	/* NOVSPRINTF */
int bufprintf(stream, fmt, arg1, arg2, arg3, arg4, arg5, arg6)
FILE *stream;
u_char *fmt;
{
	int i, len;
	u_char buf[MAXCOLS];

	len = sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5, arg6);
#endif	/* NOVSPRINTF */

	for (i = 0; i < len; i++) bufputc(buf[i], stream);
	va_end(args);
	return(len);
}

static int isoverline(c)
int c;
{
	int max;
	u_char *bollist, *eollist;

	if (kanjiflag == 2) return(0);

	max = maxcolumns;
	if (!inheader) max -= rightindent + leftindent;
	if (kanjiflag) max--;

	if (iskinsoku) {
		if (kanjiflag == 1) {
			switch (kanjimode) {
				case SJIS:
					bollist = (u_char *)KINSOKUBOLSJIS;
					eollist = (u_char *)KINSOKUEOLSJIS;
					break;
				case EUC:
					bollist = (u_char *)KINSOKUBOLEUC;
					eollist = (u_char *)KINSOKUEOLEUC;
					break;
				default:
					bollist = (u_char *)KINSOKUBOLJIS;
					eollist = (u_char *)KINSOKUEOLJIS;
					break;
			}
			if (str2chr(bollist, c, nextch)) max += 2;
			else if (str2chr(eollist, c, nextch)) max -= 2;
		}
		else {
			if (strchr((u_char *)KINSOKUBOL, c)) max++;
			else if (strchr((u_char *)KINSOKUEOL, c)) max--;
			else if (kanjimode == SJIS) {
				if (strchr((u_char *)KINSOKUBOLKANA, c)) max++;
				else if (strchr((u_char *)KINSOKUEOLKANA, c))
					max--;
			}
		}
	}
	if (column >= max) return(1);
	return(0);
}

VOID linefeed(stream)
FILE *stream;
{
	switch (returnhight) {
		case 0:	bufputc('\n', stream);
			break;
		case 1:	if (halfline) {
				halfline = 0;
				line--;
				totalline--;
			}
			else halfline = 1;
			bufprintf(stream, "\033K");
			break;
		default:
			line--;
			totalline--;
			break;
	}
	returnhight = 0;
}

VOID forcedlinefeed(c, stream)
int c;
FILE *stream;
{
	if (!inheader && maxcolumns && isoverline(c)) {
		if (kanjiflag) {
			if (linep > 2
			&& !strncmp(&linebuf[linep - 2], "\033n", 2))
				linep -= 2;
			else kanjiout(stream);
		}
		bufputc('\r', stream);
		column = 0;
		line++;
		totalline++;
		linefeed(stream);
		if (kanjiflag) kanjiin(stream);
	}
}

int foldputc(c, stream)
int c;
FILE *stream;
{
	int ret;

	if (c >= ' ') {
		forcedlinefeed(c, stream);
		if (broadsize) column++;
		else if (widesize && !isvirtical) column++;
		else if (longsize && isvirtical) column++;
		column++;
	}
	ret = bufputc(c, stream);
	if (c == '\0') {
		linebuf[linep++] = column % 0xff;
		linebuf[linep++] = column / 0xff;
	}
	if (c == '\r') column = 0;
	return(ret);
}

#ifndef	NOVSPRINTF
/*VARARGS1*/
#if	MSDOS
int foldprintf(FILE *stream, u_char *fmt, ...)
#else
int foldprintf(stream, fmt, va_alist)
FILE *stream;
u_char *fmt;
va_dcl
#endif
{
	va_list args;
	int i, len;
	u_char buf[MAXCOLS];

#if	MSDOS
	va_start(args, fmt);
#else
	va_start(args);
#endif

	len = vsprintf(buf, fmt, args);
#else	/* NOVSPRINTF */
int foldprintf(stream, fmt, arg1, arg2, arg3, arg4, arg5, arg6)
FILE *stream;
u_char *fmt;
{
	int i, len;
	u_char buf[MAXCOLS];

	len = sprintf(buf, fmt, arg1, arg2, arg3, arg4, arg5, arg6);
#endif	/* NOVSPRINTF */

	for (i = 0; i < len; i++) foldputc(buf[i], stream);
	va_end(args);
	return(len);
}

VOID printout(stream)
FILE *stream;
{
	int c;
	long count;

	kanjiflag = 0;
	kanjimode = ASCII;
	totalline = 0;
	column = line = logicalline = step = page =
	linep = halfline = leftindent = rightindent = 0;
	inheader = 0;
	if (controlchar) controlmode = 1;
	lastch = buffer = '\0';
	stackvar(1);

	count = time(0);
	nowtime = localtime(&count);

	inrange = (startpage <= 0 || page >= startpage)
		&& (endpage <= 0 || page <= endpage);

	if (inrange) header(stream);

	while ((c = fgetc(infile)) != EOF) {
		output(c, stream);
		lastch = c;
	}

	if (kanjiflag) kanjiout(stream);
	if (linep) bufputc('\r', stream);
	halfline = 0;

	if (!step && inrange) footer(stream);
	stackvar(0);
}
