/*
 *	realpath.c
 *
 *	alternative pathname parser
 */

#include "headers.h"
#include "printf.h"
#include "kctype.h"
#include "string.h"
#include "sysemu.h"
#include "pathname.h"
#include "unixemu.h"
#include "realpath.h"
#include "url.h"

#if	MSDOS && !defined (FD)
#include "termio.h"
#endif

#if	defined (FD) && (!defined (_NODOSDRIVE) || !defined (_NOFTP))
extern int reallstat __P_((CONST char *, struct stat *));
#else
#define	reallstat	Xlstat
#endif

#if	MSDOS
# ifdef	FD
extern int setcurdrv __P_((int, int));
extern VOID error __P_((CONST char *));
extern VOID lostcwd __P_((char *));
# else
static VOID NEAR error __P_((CONST char *));
# endif
#endif	/* MSDOS */

#ifdef	EMU_DOSEMU
extern char *dosgetcwd __P_((char *, int));
#endif

#ifdef	FD
extern char fullpath[];
#endif
#ifdef	EMU_DOSEMU
extern int lastdrive;
#endif

#ifndef	NOSYMLINK
static int NEAR evallink __P_((char *, int, int));
#endif
static int NEAR _realpath2 __P_((CONST char *, int, char *, int, int, int));

int norealpath = 0;


#if	MSDOS && !defined (FD)
static VOID NEAR error(s)
CONST char *s;
{
	fprintf2(stderr, "%s: Error(%d).\n", errno);
	exit(2);
}
#endif	/* MSDOS && !FD */

#ifndef	NOSYMLINK
static int NEAR evallink(path, tlen, rlen)
char *path;
int tlen, rlen;
{
	struct stat st;
	char buf[MAXPATHLEN];
	int n, duperrno;

	duperrno = errno;
	if (reallstat(path, &st) < 0) {
		errno = duperrno;
		return(-1);
	}
	if ((st.st_mode & S_IFMT) != S_IFLNK) {
		errno = duperrno;
		return(0);
	}
	if ((n = Xreadlink(path, buf, sizeof(buf) - 1)) < 0) {
		errno = duperrno;
		return(-1);
	}

	buf[n] = '\0';
	if (*buf != _SC_ && rlen) path[rlen] = '\0';
	else {
		copyrootpath(&(path[tlen]));
		rlen = tlen + 1;
	}
	rlen = _realpath2(buf, n, path, tlen, rlen, RLP_READLINK);
	errno = duperrno;

	return(rlen);
}
#endif	/* !NOSYMLINK */

/*ARGSUSED*/
static int NEAR _realpath2(path, plen, resolved, tlen, rlen, flags)
CONST char *path;
int plen;
char *resolved;
int tlen, rlen, flags;
{
#ifndef	NOSYMLINK
	int n;
#endif
	char *cp, *top;
	int len;

	if (plen <= 0) return(rlen);
	else if (path[0] != '.') /*EMPTY*/;
	else if (plen == 1) return(rlen);
	else if (path[1] != '.') /*EMPTY*/;
	else if (plen == 2) {
		top = &(resolved[tlen]);
		cp = strrdelim(top, 0);
		if (!cp || cp <= top) {
			copyrootpath(top);
			return(++tlen);
		}
		*cp = '\0';
		return(cp - top);
	}

	if ((cp = memchr2(path, _SC_, plen))) {
		len = cp++ - path;
		rlen = _realpath2(path, len, resolved, tlen, rlen, flags);
		plen -= ++len;
		return(_realpath2(cp, plen, resolved, tlen, rlen, flags));
	}

	len = snprintf2(&(resolved[rlen]), MAXPATHLEN - rlen,
		"%s%-.*s", (rlen > tlen + 1) ? _SS_ : nullstr, plen, path);
#ifndef	NOSYMLINK
	if ((flags & RLP_READLINK)
	&& (n = evallink(resolved, tlen, rlen)) > 0) {
		rlen = n;
		len = 0;
	}
#endif
	rlen += len;

	return(rlen);
}

char *realpath2(path, resolved, flags)
CONST char *path;
char *resolved;
int flags;
{
#ifdef	EMU_DOSEMU
	int duplastdrive;
#endif
#ifdef	EMU_DOSPATH
	int drv;
#endif
#if	MSDOS
	int drive;
#endif
	char *cp, tmp[MAXPATHLEN];
	int tlen, rlen;

	strcpy(tmp, path);
	path = tmp;

	cp = NULL;
	tlen = rlen = 0;
#ifdef	EMU_DOSPATH
	drv = 0;
#endif
	if (flags & RLP_PSEUDOPATH) /*EMPTY*/;
#ifdef	EMU_DOSEMU
	else if ((drv = _dospath(path))) {
		tlen = 2;
		path += tlen;
		if (*path != _SC_) {
			duplastdrive = lastdrive;
			lastdrive = drv;
			cp = dosgetcwd(resolved, MAXPATHLEN - 1);
			lastdrive = duplastdrive;
		}
		flags &= ~RLP_READLINK;
	}
#endif
#ifdef	DOUBLESLASH
	else if ((tlen = isdslash(path))) {
		rlen = snprintf2(resolved, MAXPATHLEN,
			"%-.*s%c", tlen, path, _SC_);
		path += tlen;
		flags &= ~RLP_READLINK;
	}
#endif
#ifdef	EMU_URLPATH
	else if ((tlen = urlgetscheme(path, NULL, NULL, NULL))) {
		rlen = snprintf2(resolved, MAXPATHLEN,
			"%-.*s%c", tlen, path, _SC_);
		path += tlen;
		flags &= ~RLP_READLINK;
	}
#endif
	else {
#if	MSDOS
		tlen = 2;
		drive = dospath(nullstr, NULL);
		if ((drv = _dospath(path))) path += 2;
		else drv = drive;
#else
		tlen = 0;
#endif

		if (*path == _SC_) /*EMPTY*/;
#if	MSDOS
		else if (toupper2(drv) != toupper2(drive)) {
			if (setcurdrv(drv, 0) >= 0) {
				cp = Xgetwd(resolved);
				if (setcurdrv(drive, 0) < 0)
					error("setcurdrv()");
# ifdef	FD
				if (!cp) lostcwd(resolved);
# else
				if (!cp) /*EMPTY*/;
				else
# endif
				cp = resolved;
			}
		}
#endif	/* MSDOS */
#ifdef	FD
		else if (!(flags & RLP_READLINK)
		&& resolved != fullpath && *fullpath)
			cp = strcpy(resolved, fullpath);
#endif
		else cp = Xgetwd(resolved);
	}

	if (rlen > 0) /*EMPTY*/;
	else if (cp) rlen = strlen(cp);
#ifdef	EMU_DOSPATH
	else if (drv) {
		cp = gendospath(resolved, drv, _SC_);
		rlen = cp - resolved;
	}
#endif
	else {
		copyrootpath(resolved);
		rlen = 1;
	}

	norealpath++;
	VOID_C _realpath2(path, strlen(path),
		resolved, tlen, strlen(resolved), flags);
	norealpath--;

	return(resolved);
}
